;; List Processing Library
;; microscheme.org

(define (reverse-aux lst acc)
	(if (null? lst)
		acc
		(reverse-aux (cdr lst) (cons (car lst) acc))))

(define (reverse lst)
	(reverse-aux lst ()))

(define (append l m)
	(if (null? l) m
		(cons (car l)
		(append (cdr l) m))))

(define (map-aux proc lst acc)
	(if (null? lst)
		acc
		(map-aux proc (cdr lst) (cons (proc (car lst)) acc))))

(define (map proc lst)
	(reverse-aux (map-aux proc lst ()) ()))

(define (list? x)
	(or (null? x)
		(and (pair? x) 
			(list? (cdr x)))))

(define (foldr f z xs)
	(if (null? xs)
		z
		(f (car xs) (foldr f z (cdr xs)))))

(define (foldl f z xs)
	(if (null? xs)
		z
		(foldl f (f z (car xs)) (cdr xs))))

(define (for-each proc lst)
	(if (null? lst) 
		#t
		(begin
			(proc (car lst))
			(for-each proc (cdr lst)))))

(define (fevstep proc vect i n)
	(if (< i n) 
		(begin
			(proc (vector-ref vect i))
			(fevstep proc vect (+ i 1) n))))

(define (for-each-vector proc vect)
	(fevstep proc vect 0 (vector-length vect)))
		
(define (all test lst)
	(or (null? lst)
		(and (test (car lst)) 
			(all test (cdr lst)))))

(define (vlstep vect i n acc)
	(if (< i n) 
		(vlstep vect (+ i 1) n (cons (vector-ref vect i) acc))
		acc))

(define (vector->list vect)
	(reverse (vlstep vect 0 (vector-length vect) ())))

(define (for-both p x)
	(p (car x))
	(p (cdr x)))